#include <linux/videodev2.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <ppm.h>

#include "yuyv_to_rgb.h"

#define CAMERA_DEV   "/dev/video1"
#define CAMERA_WIDTH  640
#define CAMERA_HEIGHT 480
#define MMAP_COUNT    2
#define PICTURE_NUM   30
#define PPM_FILENAME  "camera%02d.ppm"

static void ppm_writefile(uint8_t *prgb, int width, int height, int num)
{
  pixel *pixelrow;
  int i, x, y;

  pixelrow = ppm_allocrow(width);

  for (i = 0; i < num; i++) {
    char filename[16];
    FILE *fp;

    sprintf(filename, PPM_FILENAME, i);
    fp = fopen(filename, "w");

    ppm_writeppminit(fp, width, height, (pixval)255, 0);
    for (y = 0; y < height; y++) {
      for (x = 0; x < width; x++, prgb += 3)
	PPM_ASSIGN(pixelrow[x], prgb[0], prgb[1], prgb[2]);
      ppm_writeppmrow(fp, pixelrow, width, (pixval)255, 0);
    }

    fclose(fp);
  }

  ppm_freerow(pixelrow);
}

static int xioctl(int fd, int request, void *arg)
{
  for (; ; ) {
    int ret = ioctl(fd, request, arg);
    if (ret < 0) {
      if (errno == EINTR)
	continue;
      return -errno;
    }
    break;
  }

  return 0;
}

int main(void)
{
  int fd, width, height, length, count, ret, i;
  struct v4l2_format fmt;
  struct v4l2_requestbuffers req;
  struct v4l2_buffer buf;
  enum v4l2_buf_type type;
  void *mmap_p[MMAP_COUNT];
  __u32 mmap_l[MMAP_COUNT];
  uint8_t *yuyvbuf, *rgbbuf, *pyuyv, *prgb;
  struct rusage usage;
  double t;

  fd = open(CAMERA_DEV, O_RDWR, 0);
  if (fd < 0) {
    perror("open");
    return -1;
  }

  memset(&fmt, 0, sizeof(fmt));
  fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width = CAMERA_WIDTH;
  fmt.fmt.pix.height = CAMERA_HEIGHT;
  fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
  fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;
  ret = xioctl(fd, VIDIOC_S_FMT, &fmt);
  if (ret < 0 || fmt.fmt.pix.pixelformat != V4L2_PIX_FMT_YUYV ||
      fmt.fmt.pix.width <= 0 || fmt.fmt.pix.height <= 0) {
    perror("ioctl(VIDIOC_S_FMT)");
    return -1;
  }
  width = fmt.fmt.pix.width;
  height = fmt.fmt.pix.height;
  length = width * height;

  yuyvbuf = malloc(2 * length * PICTURE_NUM);
  if (!yuyvbuf) {
    perror("malloc");
    return -1;
  }

  memset(&req, 0, sizeof(req));
  req.count = MMAP_COUNT;
  req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  req.memory = V4L2_MEMORY_MMAP;
  ret = xioctl(fd, VIDIOC_REQBUFS, &req);
  if (ret < 0) {
    perror("ioctl(VIDIOC_REQBUFS)");
    return -1;
  }
  count = req.count;

  for (i = 0; i < count; i++) {
    memset(&buf, 0, sizeof(buf));
    buf.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index  = i;
    ret = xioctl(fd, VIDIOC_QUERYBUF, &buf);
    if (ret < 0) {
      perror("ioctl(VIDIOC_QUERYBUF)");
      return -1;
    }

    mmap_p[i] = mmap(NULL, buf.length, PROT_READ, MAP_SHARED, fd, buf.m.offset);
    if (mmap_p[i] == MAP_FAILED) {
      perror("mmap");
      return -1;
    }
    mmap_l[i] = buf.length;
  }

  for (i = 0; i < count; i++) {
    memset(&buf, 0, sizeof(buf));
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = i;
    ret = xioctl(fd, VIDIOC_QBUF, &buf);
    if (ret < 0) {
      perror("ioctl(VIDIOC_QBUF)");
      return -1;
    }
  }

  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  ret = xioctl(fd, VIDIOC_STREAMON, &type);
  if (ret < 0) {
    perror("ioctl(VIDIOC_STREAMON)");
    return -1;
  }

  for (i = 0, pyuyv = yuyvbuf; i < PICTURE_NUM; i++) {
    fd_set fds;

    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    for (; ; ) {
      ret = select(fd + 1, &fds, NULL, NULL, NULL);
      if (ret < 0) {
	if (errno == EINTR)
	  continue;
	perror("select");
	return -1;
      }
      break;
    }

    if (FD_ISSET(fd, &fds)) {
      memset(&buf, 0, sizeof(buf));
      buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buf.memory = V4L2_MEMORY_MMAP;
      ret = xioctl(fd, VIDIOC_DQBUF, &buf);
      if (ret < 0 || buf.bytesused < (__u32)(2 * length)) {
	perror("ioctl(VIDOC_DQBUF)");
	return -1;
      }

      memcpy(pyuyv, mmap_p[buf.index], 2 * length);
      pyuyv += 2 * length;

      ret = xioctl(fd, VIDIOC_QBUF, &buf);
      if (ret < 0) {
	perror("ioctl(VIDIOC_QBUF)");
	return -1;
      }
    }
  }

  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  xioctl(fd, VIDIOC_STREAMOFF, &type);
  for (i = 0; i < count; i++)
    munmap(mmap_p[i], mmap_l[i]);
  close(fd);

  rgbbuf = malloc(3 * length * PICTURE_NUM);
  if (!rgbbuf) {
    perror("malloc");
    return -1;
  }

  getrusage(RUSAGE_SELF, &usage);
  t = ((double)usage.ru_utime.tv_sec * 1e+3 +
       (double)usage.ru_utime.tv_usec * 1e-3);

  for (i = 0, pyuyv = yuyvbuf, prgb = rgbbuf; i < PICTURE_NUM;
       i++, pyuyv += 2 * length, prgb += 3 * length)
    yuyv_to_rgb(pyuyv, prgb, length);

  getrusage(RUSAGE_SELF, &usage);
  t = ((double)usage.ru_utime.tv_sec * 1e+3 +
       (double)usage.ru_utime.tv_usec * 1e-3) - t;
  printf("convert time: %3.3lf msec/flame\n", t / (double)PICTURE_NUM);

  free(yuyvbuf);

  ppm_writefile(rgbbuf, width, height, PICTURE_NUM);

  free(rgbbuf);

  return 0;
}
